let activeCell;

const flat = (values) => values.map((item) => item.flat ? item.flat() : item).flat();

const tryCatchParse = (value) => {
    try {
        return JSON.parse(value);
    } catch (e) {
        return value;
    }
}

const isCalculation = (value) => {
    return typeof (value) === "string" && value.match(/=.*\(/g);
}

const isCell = (value) => {
    return typeof (value) === "object" && typeof (value.address) === "object" && value.address instanceof CellAddress && typeof (value.compute) === "function";
}

const chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
const nextCol = (col) => {
    const parts = col.split(""),
        lastchar = parts[parts.length - 1];
    if (lastchar === "Z") {
        parts.push("A");
    } else {
        const nexti = chars.indexOf(lastchar) + 1;
        parts[parts.length - 1] = chars[nexti];
    }
    return parts.join();
}

const getPropertyDescriptor = (object,key) => {
    const desc = Object.getOwnPropertyDescriptor(object,key);
    if(desc || object===Object) return desc;
    return getPropertyDescriptor(Object.getPrototypeOf(object),key);
}

const AggressiveReactor = (value,cell) => {
    if(value && typeof(value)==="object" && !(value instanceof Promise)) {
        if(value.__aggressiveReactor__) return value;
        Object.defineProperty(value,"__aggressiveReactor__",{enumerable:false,value:true});
        for(const key in value) {
            const olddesc = getPropertyDescriptor(value,key)||{enumerable:true,configurable:true},
                desc = {...olddesc};
            delete desc.value;
            delete desc.writable;
            desc.set = (value) => {
                if(olddesc.set) olddesc.set(value)
                else olddesc.value = value;
                cell.compute();
                return true;
            };
            desc.get = () => {
                let value;
                if(olddesc.get) {
                    value = olddesc.get();
                    if(value && typeof(value)==="object") {
                        value = AggressiveReactor(desc.value,cell);
                    }
                } else {
                    value = olddesc.value;
                }
                if(activeCell) {
                    cell.addDependent(activeCell);
                }
                return value;
            }
            if(olddesc.value && typeof(olddesc.value)==="object") {
                olddesc.value = AggressiveReactor(olddesc.value,cell);
            }
            Object.defineProperty(value,key,desc);
        }
    }
    return value;
}

const LazyReactor = (value,cell) => {
    if(value && typeof(value)==="object" && !(value instanceof Promise)) {
        return new Proxy(value,{
            get(target,property) {
                if(activeCell) {
                    cell.addDependent(activeCell);
                }
                return LazyReactor(target[property],cell);
            },
            set(target,property,value) {
                target[property] = value;
                cell.compute();
                return true;
            }
        })
    }
    return value;
}

const Reactor = (type,value,cell) => {
    if(type===Fluxbook.VALUE_REACTIVIY_LAZY) return LazyReactor(value,cell);
    if(type===Fluxbook.VALUE_REACTIVIY_AGGRESSIVE) return AggressiveReactor(value,cell);
    return value;
}

const CellAddress = function ({sheet, row, col}) {
    if (!(this instanceof CellAddress)) return new CellAddress({sheet, row, col});
    this.sheet = sheet;
    this.row = row;
    this.col = col;
}
CellAddress.prototype.toString = function () {
    if (this.sheet) {
        if (this.row) {
            let address = `${this.sheet}!${this.col}${this.row}`;
            if(this.path) address += "." + this.path;
            return address;
        }
        return `${this.sheet}!${this.col}`;
    }
    return `${this.col}`;
}
CellAddress.prototype.toJSON = function () {
    return this.toString();
}

// body =cell.value.replace(/(?=\()?((?<sheet>[A-Z0-9]+)+(?:\!))?(?<startCol>[@_]?[A-Z]+)(?<startRow>[0-9]*)?(:?(?<endCol>[A-Z]+)(?<endRow>[0-9]*))?/gi,(match) => {

const resolveReferences = (cell) => {
    const rawArgs = [],
        body = cell.value.replace(/((?<sheet>\*|[A-Z0-9.]+)+(?:\!))?(?<startCol>[@_]?[A-Z]+)(?!\()(?<startRow>[0-9]*(\.[\._A-Z0-9]+)?)?(:?(?<endCol>[A-Z]+)(?<endRow>[0-9]*))?/gi, (match) => {
            if (cell.value.startsWith(`=${match}(`)) return match;
            rawArgs.push(match);
            return `(() => { try { return fluxbook.get("${match}",{sheet}).valueOf(); } catch(e) { console.error(e); throw new Error("Cell ${cell.address}: reference ${match} undefined in ${cell.value}\${e.message}"); }})()`;
        }).substring(1);
    return {body, rawArgs};
}

const Cell = function ({value, format, address, fluxbook} = {}) {
    const cell = {value, format, address, fluxbook, dependents: {}, listeners: {change: [], compute: []}},
        functions = {
            fluxbook() { return fluxbook?.__proxy__; },
            toJSON(_, extended) {
                const {value, format, address, computed, dependents} = cell;
                if (extended) return {
                    value,
                    format,
                    address,
                    computed,
                    dependents: Object.values(dependents).map(cell => cell.toJSON().address)
                };
                return {value, format, address};
            },
            addDependent(dependent) {
                cell.dependents[dependent.address] = dependent;
            },
            addEventListener(eventName, callback) {
                const eventnames = eventName === "*" ? Object.keys(cell.listeners) : [eventName];
                eventnames.forEach((eventName) => {
                    if (!cell.listeners[eventName].includes(callback)) {
                        cell.listeners[eventName].push(callback)
                    }
                })
            },
            compute() {
                const _activeCell = activeCell,
                    oldValue = cell.computed;
                activeCell = proxy;
                const _computed = cell.computed;
                let value;
                if(typeof(cell.value)==="function") {
                    value = cell.value.call(proxy);
                } else if (isCalculation(cell.value)) {
                    try {
                        const {body, rawArgs} = resolveReferences(cell);
                        Fluxbook.__rawArgs__ = rawArgs;
                        Fluxbook.__currentCell__ = proxy;
                        value = tryCatchParse(Function("{globalFunctions,fluxbook,sheet}", "with(globalFunctions) { return " + body + "}")({
                            globalFunctions: Fluxbook,
                            sheet: cell.address.sheet,
                            fluxbook: cell.fluxbook.__proxy__
                        }));
                        Fluxbook.__rawArgs__ = null;
                        Fluxbook.__currentCell__ = null;
                    } catch (e) {
                        value = e;
                    }
                } else {
                    value = cell.computed = cell.value;
                }
                this.withComputed(value,_computed);
                activeCell = _activeCell;
                return cell.computed;
            },
            forEachCell(callback) {
                callback(proxy);
            },
            valueOf({path}={}) {
                let value = cell.computed !== undefined ? cell.computed : cell.value;
                if(activeCell && (!value || typeof(value)!=="object" || value instanceof Promise)) {
                    this.addDependent(activeCell);
                } // else objects (which are Reactors) will handle themselves
                path ||= cell.address.path;
                if(path) {
                    const parts = path.split(".");
                    if(parts.length>0) {
                        do {
                            value = value[parts.shift()];
                        } while(parts.length>0 && value!==undefined);
                    }
                }
                return value;
            },
            withComputed(value, oldValue) {
                if (value && typeof (value) === "object" && value instanceof Promise) {
                    cell.computed = value;
                    value.then((value) => {
                        proxy.withComputed(value, oldValue);
                    })
                } else if (value !== undefined && oldValue !== value) {
                    cell.computed = value;
                    cell.listeners.compute.forEach((f) => {
                        setTimeout(f, 10, {type: "compute", target: proxy, value, oldValue})
                    })
                }
                if (oldValue !== cell.computed || (cell.computed && typeof(cell.computed)==="object")) {
                    // todo deepequal? can't just check top level value since nested props may have changed
                    if (cell.computed && typeof (cell.computed) === "object" && cell.computed instanceof Error) {
                        Object.values(cell.dependents).forEach((cell) => cell.computed = e);
                    } else {
                        Object.values(cell.dependents).forEach((cell) => {
                            cell.compute();
                        });
                    }
                }
            },
            setValue(value, {callback,path=address.path} = {}) {
                if (typeof (value) === "object" && value instanceof Promise) {
                    this.withPromise(value, {callback});
                    cell.value = value;
                } else {
                    let oldValue = cell.value;
                    if(path) {
                        const parts = path.split(".");
                        do {
                            oldValue = oldValue[parts.shift()]
                        } while(oldValue!==undefined && parts.length>0);
                    }
                    if (oldValue !== value) {
                        if(path) {
                            const parts = path.split(".");
                            let target = cell.value || Reactor(this.fluxbook().valueReactivity(),{},proxy);
                            while(parts.length>1) {
                                const key = parts.shift();
                                target = target[key] ||= {};
                            }
                            target[parts.shift()] = value;
                            // reactor will call the compute on dependencies
                        } else {
                            cell.value = Reactor(this.fluxbook().valueReactivity(),value,proxy);
                            cell.listeners.change.forEach((f) => {
                                setTimeout(f, 10, {type: "change", target: proxy, value, oldValue})
                            })
                            proxy.compute();
                        }
                    }
                }
                return cell.value;
            },
            withValue(value,{callback,path}={}) {
                this.setValue(value,{callback,path});
                return proxy;
            },
            withPromise(value, options) {
                value.__oldValue__ = cell.value;
                value.then((result) => {
                    cell.value = value.__oldValue__;
                    proxy.setValue(result, options);
                });
                return proxy;
            },
            withFormat(format) {
                cell.format = format;
                return proxy;
            },
            withEventListener(eventName, callback) {
                proxy.addEventListener(eventName, callback);
                return proxy;
            }
        },
        proxy = new Proxy(cell, {
            get(target, property) {
                return functions[property] || target[property];
            },
            set(target, property, value) {
                throw new Error(`Can't set values directly on cell!`)
            }
        });
    return proxy;
}

const Column = (config = {}) => {
    const functions = {
        toJSON(_, extended) {
            const json = {};
            Object.entries(this).forEach(([key, row]) => json[key] = row.toJSON(key, extended));
            return json;
        }
    }
    return new Proxy({}, {
        get(target, property) {
            return functions[property] || target[property];
        }
    });
}

const Sheet = ({datastore,fluxbook}={}) => {
    const functions = {
        datastore() {
            return datastore?.__proxy__ || fluxbook.datastore();
        },
        fluxbook() {
            return fluxbook;
        },
        toJSON(_, extended) {
            const json = {};
            Object.entries(this).forEach(([key, col]) => json[key] = col.toJSON(key, extended));
            return json;
        }
    }
    return new Proxy({}, {
        get(target, property) {
            return functions[property] || target[property];
        }
    });
}

const Range = ({fluxbook, sheet, startCol, endCol, startRow, endRow, path}, {create}) => {
    const original = {startCol, endCol, startRow, endRow};
    if (startCol > endCol) {
        const tempStart = startCol,
            tempend = endCol;
        endCol = tempStart;
        startCol = tempend;
    }
    if (startRow > endRow) {
        const tempStart = startCol,
            tempend = endCol;
        endRow = tempStart;
        startRow = tempend;
    }
    if (create) {
        fluxbook[sheet] ||= Sheet({fluxbook});
        fluxbook[sheet][startCol] ||= Column();
        if (endCol) {
            for (let col = nextCol(startCol); col <= endCol; col = nextCol(col)) {
                fluxbook[sheet][col] = Column();
            }
        }
        if (startRow) {
            if (endCol && endRow) {
                for (let col = startCol; col <= endCol; col = nextCol(col)) {
                    for (let row = startRow; row <= endRow; row++) {
                        fluxbook[sheet][col][row] ||= new Cell({address: CellAddress({sheet, row, col,path}), fluxbook});
                    }
                }
            } else if (endRow) {
                for (let row = startRow; row <= endRow; row++) {
                    fluxbook[sheet][startCol][row] ||= new Cell({
                        address: CellAddress({sheet, row, col: startCol,path}),
                        fluxbook
                    });
                }
            } else {
                fluxbook[sheet][startCol][startRow] ||= new Cell({
                    address: CellAddress({
                        sheet,
                        row: startRow,
                        col: startCol,
                        path
                    }), fluxbook
                });
            }
        } else if (startCol[0] === "@") {
            fluxbook[sheet][startCol] ||= new Cell({address: CellAddress({sheet, col: startCol}), fluxbook});
        }
    }
    const functions = {
            address() {
                return {fluxbook, sheet, startCol, endCol, startRow, endRow}
            },
            get(address, create) {
                if (typeof (address) === "string") {
                    const index = parseInt(address);
                    if (!isNaN(index)) return this.valueOf()[index];
                    return Ranges(fluxbook, address, {sheet, create});
                } else {
                    return [address];
                }
            },
            set(address, value, create) {
                const ranges = functions.get(address, create);
                ranges.forEach(({sheet, startCol, endCol, startRow, endRow}) => {
                    fluxbook[sheet] ||= Sheet({fluxbook});
                    fluxbook[sheet][startCol] ||= Column();
                    fluxbook[sheet][startCol][startRow] = Cell({
                        fluxbook,
                        value,
                        address: CellAddress({sheet, row: startRow, col: startCol})
                    });
                })
            },
            addDependent(dependent) {
                this.forEachCell((cell) => cell.addDependent(dependent))
            },
            contains(range, throws) {
                let ranges;
                if (typeof (range) === "string") ranges = Ranges(fluxbook, range, {sheet})
                else ranges = [range];
                return ranges.every(range => range.in({fluxbook, sheet, startCol, endCol, startRow, endRow}))
            },
            in(range, throws) {
                if (range.fluxbook !== fluxbook) return false;
                if (range.sheet !== sheet) return false;
                if (range.startCol < startCol) return false;
                if (range.startCol > endCol) return false;
                if (range.startRow < startRow) return false;
                if (range.startRow > endRow) return false;
                return true;
            },
            forEachCell(callback) {
                this.valueOf({callback, asCell: true});
            },
            withEventListener(eventName, callback) {
                const _sheet = fluxbook[sheet] ||= Sheet({datastore}),
                    column = _sheet[startCol] ||= Column({}),
                    cell = column[startRow] ||= Cell({address:CellAddress({row:startRow,col:startCol,path}),fluxbook});
                cell.withEventListener(eventName, callback);
                return proxy;
            },
            setValue(value, {callback, asCell} = {}) {
                if (startCol[0] === "@") {
                    const cell = fluxbook[sheet][startCol];
                    return cell.setValue(value,{callback,path});
                }
                if (startCol != null && endCol == null && startRow != null && endRow == null) {
                    const cell = fluxbook[sheet][startCol][startRow];
                    return cell.setValue(value,{callback,path});
                }
                if (!Array.isArray(value) || !Array.isArray(value[0])) throw new Error(`Range assignment requires an array of at least 2 dimensions`)
                let lastCol = endCol,
                    lastRow = endRow || (startRow!=null ? startRow + value.length : value.length);
                if (!endCol && startCol[0] !== "_") {
                    lastCol = startCol;
                    value.forEach((cols, i) => {
                        let nextLastCol = lastCol;
                        for(let i=0;i<cols.length;i++) nextLastCol = nextCol(nextLastCol);
                        lastCol = nextLastCol > lastCol ? nextLastCol : lastCol;
                    })
                }
                let _nextCol = nextCol;
                if (startCol[0] === "_") {
                    if (endCol && endCol !== startCol) throw new Error(`Custom column ranges must start and end with the same column`);
                    lastCol = endCol = startCol;
                    if (value.length < 2) throw new Error(`Custom column ranges must provide column names for auto-expansion due to assignment`);
                    _nextCol = col => value[0][value[0].indexOf(col)+1]; // return the next column
                    value.shift();
                    lastRow--;
                }
                const values = [];
                for (let i = 0, row = startRow || 1; i < value.length, row <= lastRow; i++, row++) {
                    const cols = value[i];
                    for (let j = 0, col = startCol; j < cols.length, col!=null && (nextCol!==_nextCol || col <= lastCol); j++, col = _nextCol(col)) {
                        let column = fluxbook[sheet][col];
                        if (!column) {
                            column = fluxbook[sheet][col] = Column();
                            endCol = col;
                        }
                        let cell = column[row];
                        if (!cell) {
                            cell = column[row] = Cell({fluxbook, address: CellAddress({sheet, row, col})});
                            endRow = row;
                        }
                        values.push(cell.setValue(cols[j],{callback,path}));
                    }
                }
                const final = {startCol, endCol, startRow, endRow};
                if (Object.entries(final).some(([key, value]) => original[key] !== value)) {
                    return fluxbook.__proxy__[`${sheet}!${final.startCol}${final.startRow}:${final.endCol}${final.endRow}`];
                }
                return values.length===0 ? values[0] : value;
            },
            flat() {
                const values = this.valueOf();
                if (Array.isArray(values)) return values.flat();
                return [values];
            },
            valueOf({callback, asCell} = {}) {
                if (!endCol && endRow == null) {
                    let cell;
                    if (startRow) {
                        cell = fluxbook[sheet]?.[startCol]?.[startRow];
                        if (cell) {
                            if(callback) callback(cell);
                            return asCell ? cell : cell.valueOf({path});
                        }
                        return;
                    }
                    cell = fluxbook[sheet]?.[startCol];
                    if (cell) {
                        if (isCell(cell)) { // is a cell
                            if (callback) callback(cell);
                            return asCell ? cell : cell.valueOf({path});
                        }
                        // rangeish
                        return Object.entries(cell).map(([key, cell]) => {
                            if (callback) callback(cell);
                            return asCell ? cell : cell.valueOf({path});
                        })
                    }
                }
                // else return 2D vertical range of columns
                const results = [];
                for (let col = startCol; col < endCol; col = nextCol(col)) {
                    const column = fluxbook[sheet][col],
                        rows = [];
                    results.push(rows);
                    let lastRow = endRow;
                    if (column) {
                        if (!endRow) {
                            lastRow = parseInt(Object.keys(column).sort((a, b) => a.length > b.length ? 1 : a > b ? 1 : a < b ? -1 : 0).pop())
                        }
                        for (let row = startRow; row <= lastRow; row++) {
                            const cell = column[row];
                            if (cell) {
                                rows[row - 1] = asCell ? cell : cell.valueOf({path});
                                if (callback) callback(cell);
                            } else {
                                rows.length = row;
                            }
                        }
                    } else {
                        rows.length = lastRow;
                    }
                }
                return results;
            }
        },
        proxy = new Proxy({}, {
            get(target, property) {
                return functions[property] || functions.get(property);
            },
            set(target, property, value) {
                functions.set(property, value, true);
                return true;
            }
        });
    return proxy;
}

const Ranges = (fluxbook, address, {sheet, create} = {}) => {
    let matchedsheet;
    const _sheet = sheet,
        functions = {
            valueOf() {
                if (ranges.length === 1) return ranges[0].valueOf();
                return ranges.map(range => range.valueOf());
            },
            forEachCell(callback) {
                ranges.forEach(range => range.forEachCell(callback));
            },
            addDependent(dependent) {
                ranges.forEach(range => range.addDependent(dependent));
            },
            withEventListener(event,listener) {
                ranges.forEach(range => range.withEventListener(event,listener));
                return ranges;
            }
        },
        ranges = [...address.matchAll(/((?<sheet>\*|[A-Z0-9.]+)+(?:\!))?(?<startCol>[@_]?[A-Z]+)(?<startRow>[0-9]*)(\.(?<path>[\._A-Z0-9]+))?(?!\()?(:?(?<endCol>[A-Z]+)(?<endRow>[0-9]*))?/gi)]
            .reduce((ranges,{groups}) => {
                let {sheet, startCol, endCol, startRow, endRow, path} = {...groups};
                if (sheet && matchedsheet !== sheet) {
                    matchedsheet = sheet;
                }
                sheet ||= matchedsheet || _sheet;
                if (sheet) {
                    sheet = sheet.toUpperCase();
                    if(sheet==="*") sheet = Object.keys(fluxbook).filter(key => key[0]!=="@").join(".");
                    sheet.split(".")
                        .forEach((sheet) => {
                            if (create) fluxbook[sheet] ||= Sheet();
                            if (endCol) endCol = endCol.toUpperCase()
                            else endCol = undefined; // sometimes ""
                            if (!startRow) {
                                if (startRow === "") startRow = undefined;  // sometimes ""
                                if (create && startCol[0] === "@") {
                                    fluxbook[sheet][startCol] ||= Cell({
                                        fluxbook,
                                        address: CellAddress({sheet, col: startCol})
                                    });
                                }
                            }
                            if (startCol) {
                                startCol = startCol.toUpperCase();
                            }
                            ranges.push(Range({
                                fluxbook,
                                sheet,
                                startCol,
                                endCol,
                                startRow: tryCatchParse(startRow),
                                endRow: tryCatchParse(endRow),
                                path
                            }, {create}));
                        })
                } else if (startCol[0] === "@") {
                    if (create) fluxbook[startCol] ||= Cell({fluxbook, address: CellAddress({col: startCol})}); // global variable
                    ranges.push(fluxbook[startCol]);
                }
                return ranges;
            },[]);
    return new Proxy(ranges, {
        get(target, property) {
            return functions[property] || target[property];
        }
    })
}

const Fluxbook = ({datastore,valueReactivity=Fluxbook.VALUE_REACTIVIY_LAZY}={}) => {
    const fluxbook = {},
        proxy = new Proxy(fluxbook, {
            get(target, property) {
                if (typeof (property) === "symbol") return;
                return functions[property] || functions.get(property);
            },
            set(target, property, value) {
                if (functions[property]) throw new Error(`${property} is a reserved word`);
                functions.set(property, value);
                return true;
            }
        }),
        functions = {
            datastore() {
                return datastore?.__proxy__;
            },
            valueReactivity() {
                return valueReactivity;
            },
            get(address, {sheet} = {}) {
                return Ranges(fluxbook, address, {sheet});
            },
            set(address, value) {
                const ranges = Ranges(fluxbook, address, {create: true});
                ranges.forEach((range) => {
                    if (isCell(range)) {
                        fluxbook[address] = range;
                    }
                    range.setValue(value);
                })
                return ranges;
            },
            toJSON(_, extended) {
                if (extended) {
                    const json = {};
                    Object.entries(fluxbook).forEach(([key, value]) => {
                        json[key] = value.toJSON(key, extended)
                    });
                    return json;
                }
                return fluxbook;
            }
        };
    Object.defineProperty(fluxbook, "__proxy__", {
        enumerable: false,
        configurable: false,
        writable: false,
        value: proxy
    });
    return proxy;
}
Fluxbook.withFunctions = (...functions) => {
    functions.forEach((f) => {
        Fluxbook[f.name] = (...args) => {
            const rawArgs = Fluxbook.__rawArgs__,
                currentCell = Fluxbook.__currentCell__;
            args = flat(args);
            if (args.some(value => typeof (value) === "object" && value instanceof Promise)) {
                return Promise.all(args).then((args) => {
                    f.rawArgs = rawArgs;
                    return f.call(currentCell,...args);
                })
            }
            f.rawArgs = rawArgs;
            return f.call(currentCell,...args);
        }
    })
    return Fluxbook;
}
Fluxbook.VALUE_REACTIVIY_NONE = 0;
Fluxbook.VALUE_REACTIVIY_LAZY = 1;
Fluxbook.VALUE_REACTIVIY_AGGRESSIVE = 2;
Object.assign(Fluxbook, {
    error(value) {
        return value;
    }
})

export {Fluxbook as default,Fluxbook};