let CURRENTOBSERVER;
const parser = new DOMParser();
const getNameFromPath = (path) => {
    const file = path.split("/").pop(),
        name = file.split(".")[0];
    if(name.includes("-")) return name;
    return "h-" + name;
}
const observe = (f,thisArg,argsList=[]) => {
    function observer(...args) {
        CURRENTOBSERVER = observer;
        try {
            f.call(thisArg||this,...argsList,...args);
        } catch(e) {

        }
        CURRENTOBSERVER = null;
    }
    observer.cancel = () => observer.cancelled = true;
    observer();
    return observer;
}
const coerce = (value,toType) => {
    const type = typeof(value);
    if(type===toType) return value;
    if(toType==="number") return parseFloat(value+"");
    if(toType==="boolean") {
        try {
            const parsed = JSON.parse(value+"");
            if(typeof(parsed)==="boolean") return parsed;
            if(parsed===1) return true;
            if(parsed===0) return false;
        } catch(e) {
            throw new TypeError(`Unable to convert ${value} into 'boolean'`);
        }
    }
    if(toType==="string") return value+"";
    const isfunction = typeof(toType)==="function";
    if((toType==="object" || isfunction) && type==="string") {
        value = value.trim();
        try {
            if(isfunction) {
                const instance = Object.create(toType.prototype);
                if(instance instanceof Array) {
                    const parsed = JSON.parse(value.startsWith("[") ? value : `[${value}]`);
                    if(!Array.isArray(parsed)) throw new TypeError(`Expected an Array for parsed data`)
                    parsed.forEach((item) => instance.push(item))
                } else {
                    Object.assign(instance,JSON.parse(value));
                }
                Object.defineProperty(instance,"constructor",{configurable:true,writable:true,value:toType.prototype.constructor||toType});
                return instance;
            }
            return JSON.parse(value);
        } catch(e) {
            throw new TypeError(`Unable to convert ${value} into ${isfunction ? toType.name : type}`);
        }
    }
    throw new TypeError(`Unable to coerce ${value} to ${toType}`)
}
const Reactor = (value) => {
    if(value && typeof(value)==="object") {
        if(value.__isReactor__) return value;
        const childReactors = [],
            dependents = {},
            proxy = new Proxy(value,{
                get(target,property) {
                    if(property==="__isReactor__") return true;
                    let value = target[property];
                    const type = typeof(value);
                    if(CURRENTOBSERVER && typeof(property)!=="symbol" && type!=="function") {
                        const observers = dependents[property] ||= new Set();
                        observers.add(CURRENTOBSERVER)
                    }
                    if(childReactors.includes(value) || (value && typeof(value)!=="object") || typeof(property)==="symbol") {
                        return value;
                    }
                    value = Reactor(value);
                    childReactors.push(value);
                    return target[property] = value;
                },
                set(target,property,value) {
                    if(target[property]!==value) {
                        if(value && typeof(value)==="object") {
                            value = Reactor(value);
                            childReactors.push(value);
                        }
                        target[property] = value;
                        const observers = dependents[property]||[];
                        [...observers].forEach((f) => {
                            if(f.cancelled) dependents[property].delete(f);
                            else f();
                        })
                    }
                    return true;
                }
            });
        return proxy;
    }
    return value;
}
const createVarsProxy = (vars,component,constructor) => {
    return new Proxy(vars,{
        get(target,property) {
            let {value} = target[property]||{};
            return value;
        },
        set(target,property,newValue) {
            if(target[property]===undefined) {
                target[property] = newValue;
                return true;
            }
            const {type,value,shared,exported} = target[property];
            if(type===undefined) {
                throw new TypeError(`Can't re-assign const ${property}`);
            }
            const newtype = typeof(newValue),
                typetype = typeof(type);
            if(type==="any" || newtype===type || (typetype==="function" && newValue && newtype==="object" && newValue instanceof type)) {
                value === newValue || (target[property].value = newValue);
                if(exported) {
                    // Array.isArray will not work here, Proxies mess up JSON.stringify for Arrays
                    if(newValue && typeof(newValue)==="object" && newValue instanceof Array) {
                        newValue = [...newValue];
                    }
                    component.setAttribute(property,typeof(newValue)==="string" ? newValue : JSON.stringify(newValue));
                }
                if(shared) {
                    constructor.instances.forEach((instance) => {
                        if(instance!==component && component.isConnected) {
                            instance.setVariable(property,newValue);
                        }
                    })
                }
                return true;
            }
            if(typetype==="function" && newValue && newtype==="object") {
                throw new TypeError(`Can't assign instance of '${newValue.constructor.name}' to variable '${property}:${type.name}'`)
            }
            throw new TypeError(`Can't assign '${typeof(newValue)} ${newtype==="string" ? '"'+newValue+'"' : newValue}' to variable '${property}:${typetype==="function" ? type.name : type}'`)
        }
    });
}
const createObserver = (domNode) => {
    const observer = new MutationObserver((mutations) => {
        mutations.forEach((mutation) => {
            if (mutation.type === "attributes") {
                const name = mutation.attributeName,
                    target = mutation.target;
                if(target.observedAttributes && target.observedAttributes.includes(name)) {
                    const value = target.getAttribute(name);
                    if(value!==mutation.oldValue) {
                        target.setVariable(name,value);
                        if(target.attributeChangedCallback) target.attributeChangedCallback(name,value,mutation.oldValue);
                    }
                }
            } else if(mutation.type==="childList") {
                for(const target of mutation.removedNodes) {
                    if(target.disconnectedCallback) target.disconnectedCallback();
                }
                for(const target of mutation.addedNodes) {
                    if(target.connectedCallback) target.connectedCallback();
                }
            }
        });
    });
    observer.observe(domNode,{subtree:true,childList:true});
    return observer;
}
const getNodes = (root) => {
    const nodes = [];
    let currentNode;
    const nodeIterator = document.createNodeIterator(root, NodeFilter.SHOW_TEXT|Node.ELEMENT_NODE|Node.DOCUMENT_FRAGMENT_NODE, {
        acceptNode(node) {
            if(node.shadowRoot) return NodeFilter.FILTER_ACCEPT;
            if(node.nodeType===Node.TEXT_NODE && node.nodeValue?.includes("${")) return NodeFilter.FILTER_ACCEPT;
            if(node.nodeType===Node.ELEMENT_NODE) {
                if(node.tagName==="A" && node.hasAttribute("href") && node.getAttribute("target")[0]==="#") {
                    return NodeFilter.FILTER_ACCEPT;
                }
                return [...node.attributes].some((attr) => {
                    return attr.nodeValue.includes("${") || attr.name.includes(":")
                }) ? NodeFilter.FILTER_ACCEPT : NodeFilter.FILTER_REJECT;
            }
            return NodeFilter.FILTER_REJECT;
        }
    });
    while (currentNode = nodeIterator.nextNode()) {
        nodes.push(currentNode);
        if (currentNode.shadowRoot) {
            nodes.push(...getNodes(currentNode.shadowRoot))
        }
    }
    return nodes;
}
const createClass = (domElementNode,observer) => {
    const instances = new Set(),
        dom = domElementNode.cloneNode(true);
    return class CustomElement extends HTMLElement {
        static get instances() {
            return instances;
        }
        constructor() {
            super();
            instances.add(this);
            observer ||= createObserver(this);
            const currentComponent = this,
                clone = dom.cloneNode(true),
                shadow = this.attachShadow({mode:"open"});
            this.vars = {
                currentComponent:{value:currentComponent},
                self:{value:currentComponent},
                boolean:{value:"boolean"},
                string:{value:"string"},
                number:{value:"number"}
            };
            this.defaultAttributes = clone.attributes;
            this.varsProxy = createVarsProxy(this.vars,this,CustomElement);
            while(clone.firstChild) shadow.appendChild(clone.firstChild);
        }
        adoptedCallback() {
            if(this.hasOwnProperty("adoptedCallback")) this.adoptedCallback();
        }
        async connectedCallback() {
            const shadow = this.shadowRoot;
            for(const attr of this.defaultAttributes) this.hasAttribute(attr.name) || this.setAttribute(attr.name,attr.value);
            for(const script of shadow.querySelectorAll("script[type=hcx]")) {
                const currentScript = document.createElement("script");
                if(script.hasAttribute("module")) script.setAttribute("type","module");
                else script.setAttribute("type","text/javascript");
                for(const attr of script.attributes) currentScript.setAttribute(attr.name,attr.value);
                const text = script.innerHTML.replaceAll(/\/\*[\s\S]*?\*\/|([^:]|^)\/\/.*$/gm,"$1").replaceAll(/\r?\n/g,"");
                currentScript.innerHTML = `Function('{ with(window.hcxContext) { ${text} }}')(); window.componentResolver(true); `;
                if(window.promisedComponent) await window.promisedComponent;
                window.hcxContext = this.varsProxy;
                window.currentScript = currentScript;
                window.promisedComponent = new Promise((resolve) => window.componentResolver = resolve);
                script.parentNode.insertBefore(currentScript,script); // wrap in try?
                script.remove();
            }
            await window.promisedComponent;
            delete window.hcxContext;
            delete window.promisedComponent;
            delete window.currentScript;
            shadow.normalize();
            const nodes = getNodes(this);
            observer.observe(this,{attributeOldValue:true});
            nodes.forEach((node) => {
                let nodes;
                if(node.nodeType===Node.TEXT_NODE) {
                    node.template = node.nodeValue;
                    nodes = [node];
                } else if(node.nodeType===Node.ELEMENT_NODE) {
                    nodes = [...node.attributes].reduce((nodes,attr) => {
                        if(attr.nodeValue.includes("${") || attr.name.includes(":")) {
                            attr.template = attr.nodeValue;
                            nodes.push(attr);
                        }
                        return nodes;
                    },[]);
                }
                nodes.forEach((node) => {
                    const observer = observe((node) => {
                        if(node.isConnected || node.ownerElement?.isConnected) {
                            if(node.template.includes("${")) {
                                node.nodeValue = Function("context","with(context) { return `" + node.template + "` }")(this.varsProxy);
                            }
                            if(node.nodeType===Node.ATTRIBUTE_NODE) {
                                const owner = node.ownerElement,
                                    value = node.nodeValue;
                                const [type,...params] = node.name.split(":");
                                if(type==="") {
                                    if(params[0]) {
                                        if(value==="true") {
                                            owner.setAttribute(params[0],"");
                                        }
                                    } else {
                                        owner.setAttribute(value,"")
                                    }
                                } else if(type==="h-on") {
                                    const event = params[0],
                                        listener = window[value] || Function(value);
                                    owner.addEventListener(event,listener);
                                } else if(type==="h-if") {
                                    owner.style.setProperty("display",value==="true" ? "revert" : "none");
                                } else if(type==="h-for") {
                                    const [what="each",vname="element",index="index",array="array",after=false] = params,
                                        coerced = coerce(value,what==="each" ? Array : "object"),
                                        target = what==="each" ? coerced : (what==="keys" ? Object.keys(coerced) : (what==="values" ? Object.values(coerced) : (what==="entries" ? Object.entries(coerced) : target))),
                                        template = owner.template ||= owner.innerHTML,
                                        html = target.reduce((html,item,i,target) => {
                                            return html += Function("context","with(context) { return `" + template + "` }")({[vname]:item,[index]:i,[array]:target})
                                        },""),
                                        parsed = parser.parseFromString(html,"text/html"),
                                        parent = owner.parentElement;
                                    if(!window.hcxDebug) {
                                        if(after) {
                                            owner.style.setProperty("display","none")
                                        } else {
                                            while(owner.lastElementChild) owner.lastElementChild.remove();
                                        }
                                    }
                                    while(parsed.body.firstChild) {
                                        if(after) {
                                            parent.insertBefore(parsed.body.firstChild,owner);
                                        } else {
                                            owner.appendChild(parsed.body.firstChild)
                                        }
                                    }
                                }
                            }
                        } else {
                            observer.cancel();
                        }
                    },this,[node]);
                })
                if(anchorImportsEnabled && node.nodeType===Node.ELEMENT_NODE && node.tagName==="A" && node.hasAttribute("href") && node.getAttribute("target")[0]==="#") {
                    node.addEventListener("click",async (event) => {
                        event.preventDefault();
                        const {as} = await importLink(node),
                            target = shadow.querySelector(node.getAttribute("target"));
                        while(target.lastElementChild) target.lastElementChild.remove();
                        target.appendChild(document.createElement(as))
                    })
                }
            })
            if(Object.hasOwnProperty("connectedCallback")) {
                this.connectedCallback();
            }
        }
        adopted(value) {
            Object.defineProperty(this,"adoptedCallback",{configurable:true,writable:true,value});
        }
        connected(value) {
            Object.defineProperty(this,"connectedCallback",{configurable:true,writable:true,value});
        }
        attributeChanged(value) {
            Object.defineProperty(this,"attributeChangedCallback",{configurable:true,writable:true,value});
        }
        disconnected(value) {
            Object.defineProperty(this,"disconnectedCallback",{configurable:true,writable:true,value});
        }
        setVariable(name,value,coerceTo=typeof(value)) {
            let {type,...rest} = this.vars[name]||{};
            if(type) {
                const oldValue = this.varsProxy[name];
                value =  coerce(value,type);
                if(oldValue!==value) this.varsProxy[name] = value;
                return true;
            }
            this.vars[name] = {type,value:coerce(value,coerceTo)};
            return false;
        }
        getVariable(name) {
            return this.varsProxy[name];
        }
        variables(variables,options={}) { // options = {observed,reactive,shared,exported,imported}
            if(variables!==undefined) {
                Object.entries(variables)
                    .forEach(([key,type]) => {
                        const variable =  this.vars[key];
                        if(variable===undefined) {
                            this.vars[key] = {type};
                        }
                    })
            }
            Object.entries(options).forEach(([key,bool]) => {
                if(bool) {
                    this[key](variables,true)
                }
            });
            return Object.entries(this.vars)
                .reduce((result,[key,variable]) => {
                    result[key] = {...variable};
                    return result;
                },{})
        }
        observed(variables,exist) {
            if(!exist) this.variables(variables);
            Object.entries(variables)
                .forEach(([key]) => {
                    const variable =  this.vars[key];
                    variable.value = coerce(this.getAttribute(key),variable.type);
                    variable.observed = true;
                })
            const result = this.variables();
            Object.entries(result).forEach((key,value) => value.observed || (delete result[key]));
            Object.defineProperty(this,"observedAttributes",{configurable:true,writable:true,value:Object.keys(result)});
            return result;
        }
        reactive(variables,exist) {
            if(!exist) this.variables(variables);
            Object.entries(variables)
                .forEach(([key]) => {
                    const variable =  this.vars[key];
                    variable.reactive = true;
                    this.vars[key]  = Reactor(variable);
                });
            const result = this.variables();
            Object.entries(result).forEach((key,value) => value.reactive || (delete result[key]));
            return result;
        }
        shared(variables,exist) {
            if(!exist) this.variables(variables);
            Object.entries(variables)
                .forEach(([key]) => {
                    const variable =  this.vars[key];
                    variable.shared = true;
                });
            const result = this.variables();
            Object.entries(result).forEach((key,value) => value.shared || (delete result[key]));
            return result;
        }
        exported(variables,exist) {
            if(!exist) this.variables(variables);
            Object.entries(variables)
                .forEach(([key]) => {
                    const variable =  this.vars[key];
                    variable.exported = true;
                });
            const result = this.variables();
            Object.entries(result).forEach((key,value) => value.exported || (delete result[key]));
            return result;
        }
        imported(variables,exist) {
            if(!variables) {
                variables = [...this.attributes].reduce((variables,{name}) => { variables[name] = "string"; return variables; },{});
            }
            if(!exist) this.variables(variables);
            Object.entries(variables)
                .forEach(([key,type]) => {
                    const variable =  this.vars[key];
                    variable.value = coerce(this.getAttribute(key),type);
                    variable.imported = true;
                });
            const result = this.variables();
            Object.entries(result).forEach((key,value) => value.exported || (delete result[key]));
            return result;
        }
    }
}
const createComponent = (name,node,observer)  => {
    customElements.define(name,createClass(node,observer));
}
const importLink = async (link,observer) => {
    const url = (new URL(link.getAttribute("href"),window.location.href)),
        as = link.getAttribute("as") || getNameFromPath(url.pathname);
    if(url.hostname!==window.location.hostname) {
        throw new URIError(`importLink:HTML imports must be from same domain: ${url.hostname}!=${location.hostname}`)
    }
    if(!customElements.get(as)) {
        const html = await (await fetch(url.href)).text(),
            dom = parser.parseFromString(html,"text/html");
        createComponent(as,dom.body,observer);
    }
    return {as};
}
const importComponents = async () => {
    const observer = createObserver(document.body);
    for (const link of [...document.querySelectorAll("link[href][rel=module]")]) {
        await importLink(link);
    }
}
let anchorImportsEnabled;
const enableAnchorImports = () => {
    anchorImportsEnabled = true;
}

const bodyAsComponent = (as) => {
    createComponent(as,document.body);
    const component = document.createElement(as);
    document.querySelector("html").replaceChild(component,document.body);
}
const url = new URL(import.meta.url);
if(url.searchParams.has("as")) {
    if(url.searchParams.get("anchorImports")==="true") {
        enableAnchorImports();
    }
    bodyAsComponent(url.searchParams.get("as"));
}

export {importComponents,enableAnchorImports}